@@ -1,3 +1,5 @@
+--- @diagnostic disable:no-unknown
+
 --- @class vim._inspector.Filter
 --- @inlinedoc
 ---
@@ -27,6 +29,7 @@ local defaults = {
 ---
 ---Can also be pretty-printed with `:Inspect!`. [:Inspect!]()
 ---
+---@since 11
 ---@param bufnr? integer defaults to the current buffer
 ---@param row? integer row to inspect, 0-based. Defaults to the row of the current cursor
 ---@param col? integer col to inspect, 0-based. Defaults to the col of the current cursor
@@ -52,7 +55,7 @@ function vim.inspect_pos(bufnr, row, col, filter)
     local cursor = vim.api.nvim_win_get_cursor(win)
     row, col = cursor[1] - 1, cursor[2]
   end
-  bufnr = bufnr == 0 and vim.api.nvim_get_current_buf() or bufnr
+  bufnr = vim._resolve_bufnr(bufnr)
 
   local results = {
     treesitter = {}, --- @type table[]
@@ -77,6 +80,7 @@ function vim.inspect_pos(bufnr, row, col, filter)
   -- treesitter
   if filter.treesitter then
     for _, capture in pairs(vim.treesitter.get_captures_at_pos(bufnr, row, col)) do
+      --- @diagnostic disable-next-line: inject-field
       capture.hl_group = '@' .. capture.capture .. '.' .. capture.lang
       results.treesitter[#results.treesitter + 1] = resolve_hl(capture)
     end
@@ -84,7 +88,7 @@ function vim.inspect_pos(bufnr, row, col, filter)
 
   -- syntax
   if filter.syntax and vim.api.nvim_buf_is_valid(bufnr) then
-    vim.api.nvim_buf_call(bufnr, function()
+    vim._with({ buf = bufnr }, function()
       for _, i1 in ipairs(vim.fn.synstack(row + 1, col + 1)) do
         results.syntax[#results.syntax + 1] =
           resolve_hl({ hl_group = vim.fn.synIDattr(i1, 'name') })
@@ -100,40 +104,42 @@ function vim.inspect_pos(bufnr, row, col, filter)
 
   --- Convert an extmark tuple into a table
   local function to_map(extmark)
-    extmark = {
+    local opts = resolve_hl(extmark[4])
+    return {
       id = extmark[1],
       row = extmark[2],
       col = extmark[3],
-      opts = resolve_hl(extmark[4]),
+      end_row = opts.end_row or extmark[2],
+      end_col = opts.end_col or extmark[3],
+      opts = opts,
+      ns_id = opts.ns_id,
+      ns = nsmap[opts.ns_id] or '',
     }
-    extmark.ns_id = extmark.opts.ns_id
-    extmark.ns = nsmap[extmark.ns_id] or ''
-    extmark.end_row = extmark.opts.end_row or extmark.row -- inclusive
-    extmark.end_col = extmark.opts.end_col or (extmark.col + 1) -- exclusive
-    return extmark
   end
 
-  --- Check if an extmark overlaps this position
-  local function is_here(extmark)
-    return (row >= extmark.row and row <= extmark.end_row) -- within the rows of the extmark
-      and (row > extmark.row or col >= extmark.col) -- either not the first row, or in range of the col
-      and (row < extmark.end_row or col < extmark.end_col) -- either not in the last row or in range of the col
+  --- Exclude end_col and unpaired marks from the overlapping marks, unless
+  --- filter.extmarks == 'all' (a highlight is drawn until end_col - 1).
+  local function exclude_end_col(extmark)
+    return filter.extmarks == 'all' or row < extmark.end_row or col < extmark.end_col
   end
 
-  -- all extmarks at this position
-  local extmarks = vim.api.nvim_buf_get_extmarks(bufnr, -1, 0, -1, { details = true })
+  -- All overlapping extmarks at this position:
+  local extmarks = vim.api.nvim_buf_get_extmarks(bufnr, -1, { row, col }, { row, col }, {
+    details = true,
+    overlap = true,
+  })
   extmarks = vim.tbl_map(to_map, extmarks)
-  extmarks = vim.tbl_filter(is_here, extmarks)
+  extmarks = vim.tbl_filter(exclude_end_col, extmarks)
 
   if filter.semantic_tokens then
     results.semantic_tokens = vim.tbl_filter(function(extmark)
-      return extmark.ns:find('vim_lsp_semantic_tokens') == 1
+      return extmark.ns:find('nvim.lsp.semantic_tokens') == 1
     end, extmarks)
   end
 
   if filter.extmarks then
     results.extmarks = vim.tbl_filter(function(extmark)
-      return extmark.ns:find('vim_lsp_semantic_tokens') ~= 1
+      return extmark.ns:find('nvim.lsp.semantic_tokens') ~= 1
         and (filter.extmarks == 'all' or extmark.opts.hl_group)
     end, extmarks)
   end
@@ -145,6 +151,14 @@ end
 ---
 ---Can also be shown with `:Inspect`. [:Inspect]()
 ---
+---Example: To bind this function to the vim-scriptease
+---inspired `zS` in Normal mode:
+---
+---```lua
+---vim.keymap.set('n', 'zS', vim.show_pos)
+---```
+---
+---@since 11
 ---@param bufnr? integer defaults to the current buffer
 ---@param row? integer row to inspect, 0-based. Defaults to the row of the current cursor
 ---@param col? integer col to inspect, 0-based. Defaults to the col of the current cursor
@@ -169,7 +183,7 @@ function vim.show_pos(bufnr, row, col, filter)
     if data.hl_group ~= data.hl_group_link then
       append('links to ', 'MoreMsg')
       append(data.hl_group_link, data.hl_group_link)
-      append(' ')
+      append('   ')
     end
     if comment then
       append(comment, 'Comment')
@@ -182,7 +196,16 @@ function vim.show_pos(bufnr, row, col, filter)
     append('Treesitter', 'Title')
     nl()
     for _, capture in ipairs(items.treesitter) do
-      item(capture, capture.lang)
+      item(
+        capture,
+        string.format(
+          'priority: %d   language: %s',
+          capture.metadata.priority
+            or (capture.metadata[capture.id] and capture.metadata[capture.id].priority)
+            or vim.hl.priorities.treesitter,
+          capture.lang
+        )
+      )
     end
     nl()
   end
