@@ -127,6 +127,7 @@
   "eval"
   "sign"
   "abort"
+  "substitute"
 ] @keyword
 
 (map_statement
@@ -252,6 +253,9 @@
 (heredoc
   (parameter) @keyword)
 
+(script
+  (parameter) @keyword)
+
 [
   (marker_definition)
   (endmarker)
@@ -287,6 +291,7 @@
   "=~"
   "!~"
   "="
+  "^="
   "+="
   "-="
   "*="
@@ -312,6 +317,9 @@
 (binary_operation
   "." @operator)
 
+(lua_statement
+  "=" @keyword)
+
 ; Punctuation
 [
   "("
